@@ -331,12 +331,11 @@ typedef Label *Xt;
 #define VTLINK 0
 #define VTCOMPILE 1
 #define VTTO 2
-#define VTDEFER 3
-#define VTEXTRA 4
-#define VT2INT 5
-#define VT2COMP 6
-#define VT2STRING 7
-#define VT2LINK 8
+#define VTEXTRA 3
+#define VT2INT 4
+#define VT2COMP 5
+#define VT2STRING 6
+#define VT2LINK 7
 #define EXTRA_CODE(cfa) ((Xt *)(((Cell **)cfa)[-1][VTEXTRA]))
 #define EXTRA_CODEXT(cfa) ((Xt)(((Cell **)cfa)[-1][VTEXTRA]))
 
@@ -506,17 +505,16 @@ int gforth_args(int argc, char ** argv, char ** path, char ** imagename);
 ImageHeader* gforth_loader(char* imagename, char* path);
 user_area* gforth_stacks(Cell dsize, Cell rsize, Cell fsize, Cell lsize);
 void gforth_free_stacks(user_area* t);
-void gforth_free();
+void gforth_free(void * ptr);
 Cell gforth_go(Xt* ip0);
 Cell gforth_boot(int argc, char** argv, char* path);
-void gforth_bootmessage();
+Cell gforth_bootmessage();
 Cell gforth_start(int argc, char ** argv);
 Cell gforth_quit();
 Xt gforth_find(Char * name);
 Cell gforth_execute(Xt xt);
 void gforth_cleanup();
 void gforth_printmetrics();
-void gforth_setwinch();
 void gforth_setstacks(user_area * t);
 #if defined(DOUBLY_INDIRECT)
 Cell gforth_make_image(int debugflag);
@@ -524,7 +522,6 @@ Cell gforth_make_image(int debugflag);
 #ifdef HAVE_MCHECK
 void gforth_abortmcheck(enum mcheck_status reason);
 #endif
-void gforth_free(void * ptr);
 void gforth_sigset(sigset_t *set, ...);
 
 #if SIZEOF_CHAR_P == 4
@@ -607,6 +604,8 @@ UDCell dadd(UDCell x, UDCell y);
 
 Cell memcasecmp(const Char *s1, const Char *s2, Cell n);
 
+DCell utf8_fetch_plus(Char * c_addr, UCell len);
+
 void vm_print_profile(FILE *file);
 void vm_count_block(Xt *ip);
 
