@@ -187,12 +187,18 @@ Address code_area=0;
 Cell code_area_size = CODE_BLOCK_SIZE;
 Address code_here; /* does for code-area what HERE does for the dictionary */
 Address start_flush=NULL; /* start of unflushed code */
-Cell last_jump=0; /* if the last prim was compiled without jump, this
-                     is it's number, otherwise this contains 0 */
+PrimNum last_jump=0; /* if the last prim was compiled without jump, this
+                        is it's PrimNum, otherwise this contains 0 */
+Cell ip_at=0; /* ip currently points to the prim at ip_at */
+#define MAX_IP_UPDATE 16
+Cell inst_index; /* current instruction */
+Label **ginstps; /* array of threaded code locations for
+                           primitives being optimize_rewrite()d */
 
 static int no_super=0;   /* true if compile_prim should not fuse prims */
 static int no_dynamic=NO_DYNAMIC_DEFAULT; /* if true, no code is generated
 					     dynamically */
+static int no_dynamic_image=0; /* disable dynamic while loading the image */
 static int no_rc0=0;  /* true if don't load ~/.config/gforthrc0 */
 static int print_metrics=0; /* if true, print metrics on exit */
 static int print_prims=0; /* if true, print primitives on exit */
@@ -200,6 +206,13 @@ static int static_super_number = 10000; /* number of ss used if available */
 #define MAX_STATE 9 /* maximum number of states */
 #define CANONICAL_STATE 0
 static int maxstates = MAX_STATE; /* number of states for stack caching */
+static int opt_ip_updates =  /* 0=disable, 1=simple, >=2=also optimize ;s */
+#ifdef GFORTH_DEBUGGING
+  0
+#else
+  100
+#endif
+  ;
 static int ss_greedy = 0; /* if true: use greedy, not optimal ss selection */
 static int tpa_noequiv = 0;     /* if true: no state equivalence checking */
 static int tpa_noautomaton = 0; /* if true: no tree parsing automaton */
@@ -235,7 +248,8 @@ Label *labels; /* labels, as pointed to by vm_prims */
 
 typedef struct {
   Label start; /* NULL if not relocatable */
-  Cell length; /* only includes the jump iff superend is true*/
+  Cell len1;   /* the length of the ip update */
+  Cell length; /* includes len1; only includes the jump iff superend is true*/
   Cell restlength; /* length of the rest (i.e., the jump or (on superend) 0) */
   unsigned uses; /* number of uses */
   char superend; /* true if primitive ends superinstruction, i.e.,
@@ -357,7 +371,7 @@ static unsigned char *branch_targets(Cell *image, const unsigned char *bitstring
   for(k=0; k<steps; k++) {
     for(j=0, bits=bitstring[k]; j<RELINFOBITS; j++, i++, bits<<=1) {
       if(bits & (1U << (RELINFOBITS-1))) {
-	assert(i*sizeof(Cell) < size);
+	assert(i < steps*RELINFOBITS);
         token=image[i];
 	if ((token>=base) &&
 	    (SECTION(token) == sect)) { /* relocatable address */
@@ -419,7 +433,7 @@ void gforth_relocate(Address sections[], Char *bitstrings[],
 	/*      fprintf(stderr,"relocate: image[%d]\n", i);*/
 	if(bits & (1U << (RELINFOBITS-1))) {
 	  // debugp(stderr,"relocate: image[%d]=%d of %d\n", i, image[i], size/sizeof(Cell));
-	  assert(i*sizeof(Cell) < size);
+	  assert(i < steps*RELINFOBITS);
 	  token=image[i];
 	  if(SECTION(token)==0xFF) {
 	    int group = (-token & 0x3E00) >> 9;
@@ -428,7 +442,9 @@ void gforth_relocate(Address sections[], Char *bitstrings[],
 	      case CF_NIL      : image[i]=0; break;
 #if !defined(DOUBLY_INDIRECT)
 	      case CF(DOER_MAX) ... CF(DOCOL):
-		MAKE_CF(image+i,symbols[CF(token)]); break;
+		compile_prim1(0); /* flush primitive state whatever it is in */
+		MAKE_CF(image+i,symbols[CF(token)]);
+		break;
 #endif /* !defined(DOUBLY_INDIRECT) */
 	      default          : /* backward compatibility */
 		/*	      printf("Code field generation image[%x]:=CFA(%x)\n",
@@ -810,7 +826,8 @@ struct cost { /* super_info might be a more accurate name */
   char branch;	    /* is it a branch (SET_IP) */
   unsigned char state_in;    /* state on entry */
   unsigned char state_out;   /* state on exit */
-  unsigned char imm_ops;     /* number of immediate operands */
+  unsigned char imm_ops;     /* number of additional threaded-code slots
+                                (immediate arguments+number of components-1) */
   short offset;     /* offset into super2 table */
   unsigned char length;      /* number of components */
 };
@@ -983,10 +1000,10 @@ static void gforth_printprims()
   for (i=0; i<npriminfos; i++) {
     PrimInfo *pi=&priminfos[i];
     struct cost *sc=&super_costs[i];
-    fprintf(stderr,"%-15s %d-%d %4d %4d %12p len=%3ld rest=%2ld send=%1d\n",
+    fprintf(stderr,"%-17s %d-%d %4d %4d %14p len=%2ld+%3ld+%2ld send=%1d\n",
 	    prim_names[i], state_map(sc->state_in), state_map(sc->state_out),
-	    i, pi->uses, pi->start, (long)(pi->length), (long)(pi->restlength),
-	    pi->superend);
+	    i, pi->uses, pi->start, (long)(pi->len1),
+            (long)(pi->length - pi->len1), (long)(pi->restlength), pi->superend);
   }
 }
 #endif
@@ -995,7 +1012,10 @@ static void check_prims(Label symbols1[])
 {
   int i;
 #ifndef NO_DYNAMIC
-  Label *symbols2, *ends1, *ends1j, *ends1jsorted, *goto_p;
+  Label *symbols2, *goto_p;
+  Label *startb;                /* L labels (LABEL1 right behind ip update)*/
+  Label *ends1;                 /* K labels (LABEL2 right before NEXT_P1_5) */
+  Label *ends1j, *ends1jsorted; /* J labels (LABEL3 right before DO_GOTO) */
   int nends1j;
 #endif
 
@@ -1015,7 +1035,8 @@ static void check_prims(Label symbols1[])
   if (no_dynamic)
     return;
   symbols2=gforth_engine2(0 sr_call);
-  ends1 = symbols1+i+1;
+  startb = symbols1+i+1;
+  ends1  = startb+i;
   ends1j =   ends1+i;
   goto_p = ends1j+i+1; /* goto_p[0]==before; ...[1]==after;*/
   nends1j = i+1;
@@ -1048,6 +1069,7 @@ static void check_prims(Label symbols1[])
     Label endlabel = bsearch_next(symbols1[i]+1,ends1jsorted,nends1j);
 
     pi->start = s1;
+    pi->len1 = ((char *)startb[i])-s1;
     pi->superend = superend[i]|no_super;
     pi->length = prim_len;
     pi->restlength = endlabel - symbols1[i] - pi->length;
@@ -1063,10 +1085,10 @@ static void check_prims(Label symbols1[])
 	debugp(stderr, "S%d: op%d(S%d) = %d (%d);", sc->state_in, p, sc->state_out, i+1, pi->length);
     }
 #else
-    debugp(stderr, "%-15s %d-%d %4d %p %p len=%3ld rest=%2ld send=%1d",
+    debugp(stderr, "%-15s %d-%d %4d %p %p len=%2ld+%3ld+%2ld send=%1d",
 	   prim_names[i], state_map(sc->state_in), state_map(sc->state_out),
-	   i, s1, s2, (long)(pi->length), (long)(pi->restlength),
-	   pi->superend);
+	   i, s1, s2, (long)(pi->len1),
+           (long)(pi->length - pi->len1), (long)(pi->restlength), pi->superend);
 #endif
     if (endlabel == NULL) {
       pi->start = NULL; /* not relocatable */
@@ -1217,12 +1239,53 @@ static void MAYBE_UNUSED align_code(void)
 #endif /* defined(NO_DYNAMIC */
 }  
 
+static Cell *ip_update_metrics = NULL;
+static Cell nip_update_metrics=0;
+
 #ifndef NO_DYNAMIC
+static void record_ip_update(Cell n)
+{
+  if (n>=nip_update_metrics) {
+    ip_update_metrics = realloc(ip_update_metrics, (n+1)*sizeof(Cell));
+    memset(ip_update_metrics+nip_update_metrics,0,
+           (n+1-nip_update_metrics)*sizeof(Cell));
+    nip_update_metrics=n+1;
+  }
+  ip_update_metrics[n]++;
+}
+
+static void append_ip_update()
+/* compile an ip update for updating the ip from ip_at to
+   inst_index+1, where both are indexes into ginstps */
+{
+  /* assert(ip_at <= inst_index+1); */
+  if (ip_at < inst_index+1) {
+    Cell cellsdiff = ginstps[inst_index+1]-ginstps[ip_at];
+    assert(opt_ip_updates > 0);
+    if (print_metrics)
+      record_ip_update(cellsdiff);
+    do {
+      Cell cellsdiff1 = cellsdiff;
+      if (cellsdiff1 > MAX_IP_UPDATE)
+        cellsdiff1 = MAX_IP_UPDATE;
+      {
+        PrimNum p = N_noop-1+cellsdiff1;
+        PrimInfo *pi = &priminfos[p];
+        append_code(pi->start, pi->len1);
+      }
+      cellsdiff -= cellsdiff1;
+    } while (cellsdiff>0);
+    ip_at = inst_index+1;
+  }
+}
+
 static void append_jump(void)
 {
   if (last_jump) {
     PrimInfo *pi = &priminfos[last_jump];
     /* debugp(stderr, "Copy code %p<=%p+%x,%d\n", code_here, pi->start, pi->length, pi->restlength); */
+    assert(ip_at > inst_index || !priminfos[last_jump].superend);
+    append_ip_update();
     append_code(pi->start+pi->length, pi->restlength);
     /* debugp(stderr, "Copy goto %p<=%p,%d\n", code_here, goto_start, goto_len); */
     append_code(goto_start, goto_len);
@@ -1234,7 +1297,9 @@ static void append_jump(void)
 static void append_jump_previous(void)
 /* append a dispatch to the previous primitive */
 {
+  inst_index--;
   append_jump();
+  inst_index++;
 }
 
 /* Gforth remembers all code blocks in this list.  On forgetting (by
@@ -1279,15 +1344,39 @@ static int reserve_code_space(UCell size)
   return 0;
 }
 
-static Address append_prim(Cell p)
+/* primitives, where ip is dead at the start, so no ip update is needed */
+static PrimNum ip_dead[] = {N_semis, N_execute_semis, N_fast_throw};
+
+static Address append_prim(PrimNum p)
 {
   PrimInfo *pi = &priminfos[p];
+  struct cost *ci = &super_costs[p];
   Address old_code_here;
   if(reserve_code_space(pi->length+pi->restlength+goto_len+CODE_ALIGNMENT-1))
     return NULL;
-  /* debugp(stderr, "Copy code %p<=%p,%d\n", code_here, pi->start, pi->length); */
+  /* debugp(stderr, "Copy code %p<=%p,%d\n", code_here, pi->start, pi->length);*/
   old_code_here = code_here;
-  append_code(pi->start, pi->length);
+  if (opt_ip_updates>0) {
+    if (ci->imm_ops+1>ci->length || pi->superend) {
+      inst_index += ci->length-1; /* -1 to correct for the +1 in: */
+      if (opt_ip_updates>1) {
+        long i;
+        for (i=0; i<sizeof(ip_dead)/sizeof(ip_dead[0]); i++)
+          if (p==ip_dead[i]) {
+            ip_at = inst_index+1; /* suppress the ip update if ip is dead */
+            break;
+          }
+      }
+      append_ip_update();
+    }
+    append_code(pi->start+pi->len1, pi->length-pi->len1);
+  } else {
+    if (print_metrics)
+      record_ip_update(ci->imm_ops+1);
+    append_code(pi->start, pi->length);
+    ip_at = inst_index + ci->length;
+  }
+  last_jump = (pi->restlength == 0) ? 0 : p;
   return old_code_here;
 }
 
@@ -1414,7 +1503,7 @@ void finish_code_barrier(void)
 }
 
 #if !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED))
-static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
+static Cell compile_prim_dyn(PrimNum p)
      /* compile prim #p dynamically (mod flags etc.) and return start
         address of generated code for putting it into the threaded code */
 {
@@ -1430,6 +1519,7 @@ static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
   priminfos[p].uses++;
   if (p>=npriminfos || !is_relocatable(p)) {
     append_jump_previous();
+    ip_at = inst_index+1; /* advance to behind the non-relocatable inst */
     return static_prim;
   }
   old_code_here = append_prim(p);
@@ -1444,7 +1534,11 @@ static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
 #ifndef NO_DYNAMIC
 static int cost_codesize(int prim)
 {
-  return priminfos[prim].length;
+  PrimInfo *pi = &priminfos[prim];
+  int cost =  pi->length;
+  if (opt_ip_updates>0)
+    cost -= pi->len1;
+  return cost;
 }
 #endif
 
@@ -1699,9 +1793,16 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 {
   int i,j;
   struct tpa_state *ts[ninsts+1];
-  int nextdyn, nextstate, startstate, no_transition;
+  int nextdyn, nextstate, startstate, no_transition, no_relocatable=0;
   Address old_code_area;
+  ginstps = (Label **)instps;
   DynamicInfo *di = NULL;
+  Cell last=origs[ninsts-1];
+  struct cost *cl=&super_costs[last];
+  if (ninsts == 0)
+    return;
+  instps[ninsts] = instps[ninsts-1]+cl->imm_ops+1;
+  /* !! better provide this info through the interface */
   lb_basic_blocks++;
   ts[ninsts] = termstate;
 #ifndef NO_DYNAMIC
@@ -1768,10 +1869,12 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
   old_code_area = code_area;
   nextdyn=0;
   nextstate=CANONICAL_STATE;
+  ip_at = 0;
   no_transition = ((!ts[0]->trans[nextstate].relocatable) 
 		   ||ts[0]->trans[nextstate].no_transition);
   for (i=0; i<ninsts; i++) {
     Cell tc=0, tc2;
+    inst_index=i;
     startstate = nextstate;
     if (i==nextdyn) {
       PrimNum p;
@@ -1781,7 +1884,7 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 	struct cost *c = super_costs+pt;
 	assert(ts[i]->trans[nextstate].cost != INF_COST);
 	assert(c->state_in==nextstate);
-	tc = compile_prim_dyn(pt,NULL);
+	tc = compile_prim_dyn(pt);
 	nextstate = c->state_out;
       }
       p = ts[i]->inst[nextstate].inst;
@@ -1793,8 +1896,9 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 #if defined(GFORTH_DEBUGGING)
 	assert(p == origs[i]);
 #endif
-	tc2 = compile_prim_dyn(p,instps[i]);
-	if (no_transition || !is_relocatable(p)) {
+	tc2 = compile_prim_dyn(p);
+	no_relocatable = !is_relocatable(p);
+	if (no_transition || no_relocatable) {
 	  /* !! actually what we care about is if and where
 	   * compile_prim_dyn() puts NEXTs */
 	  tc=tc2;
@@ -1803,6 +1907,10 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 	no_transition = ts[i]->inst[nextstate].no_transition;
 	nextstate = c->state_out;
 	nextdyn += c->length;
+	/* if the starter of a i<nextdyn sequence was non-relocatable,
+	 * the rest of the sequence needs to keep up ip_at */
+	if(no_relocatable)
+	  ip_at = nextdyn;
       }
       if (is_relocatable(p)) {
         di = add_dynamic_info();
@@ -1825,14 +1933,17 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
       tc= (Cell)vm_prims[super2[super_costs[ts[i]->inst[CANONICAL_STATE].inst].offset]];
     }
     *(instps[i]) = tc;
-  }      
+  }
+  assert(inst_index == i-1);
+  if (!no_dynamic)
+    append_ip_update();
   if (!no_transition) {
     PrimNum p = ts[i]->trans[nextstate].inst;
     struct cost *c = super_costs+p;
     assert(c->state_in==nextstate);
     assert(ts[i]->trans[nextstate].cost != INF_COST);
     assert(i==nextdyn);
-    (void)compile_prim_dyn(p,NULL);
+    (void)compile_prim_dyn(p);
     nextstate = c->state_out;
     di->length = code_here - (Address)di->start;
     di->end_state = nextstate;
@@ -1863,8 +1974,8 @@ void compile_prim1(Cell *start)
 #elif defined(INDIRECT_THREADED)
   return;
 #else /* !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED)) */
-  static Cell *instps[MAX_BB];
-  static PrimNum origs[MAX_BB];
+  static Cell *instps[MAX_BB+1];
+  static PrimNum origs[MAX_BB+1];
   static int ninsts=0;
   PrimNum prim_num;
 
@@ -2105,6 +2216,7 @@ ImageHeader* gforth_loader(char* imagename, char* path)
   Cell data_offset = offset_image ? 56*sizeof(Cell) : 0;
   UCell check_sum;
   FILE* imagefile=open_image_file(imagename, path);
+  int no_dynamic_orig = no_dynamic;
 
   if(imagefile == NULL) return NULL;
   if(gforth_init()) return NULL;
@@ -2189,7 +2301,9 @@ ImageHeader* gforth_loader(char* imagename, char* path)
     debugp(stderr, "section base=%p, dp=%p, size=%lx\n", section.base, section.dp, section.size);
     if(fread(sections[i], 1, sizes[i], imagefile) != sizes[i]) break;
   }
+  no_dynamic |= no_dynamic_image;
   gforth_relocate(sections, reloc_bits, sizes, bases, vm_prims);
+  no_dynamic = no_dynamic_orig;
 #if 0
   { /* let's see what the relocator did */
     FILE *snapshot=fopen("snapshot.fi","wb");
@@ -2280,6 +2394,7 @@ enum {
   ss_min_lsu,
   ss_min_nexts,
   opt_code_block_size,
+  opt_opt_ip_updates,
 };
 
 static void print_diag()
@@ -2303,7 +2418,7 @@ static void print_diag()
 #endif
      )
     debugp(stderr, "relocs: %d:%d\n", relocs, nonrelocs);
-    fprintf(stderr, "*** %sperformance problems ***\n%s%s",
+  fprintf(stderr, "*** %sperformance problems ***\n%s%s",
 #if defined(BUGGY_LL_CMP) || defined(BUGGY_LL_MUL) || defined(BUGGY_LL_DIV) || defined(BUGGY_LL_ADD) || defined(BUGGY_LL_SHIFT) || defined(BUGGY_LL_D2F) || defined(BUGGY_LL_F2D) || !(defined(FORCE_REG) || defined(FORCE_REG_UNNECESSARY)) || defined(BUGGY_LONG_LONG) || (NO_DYNAMIC_DEFAULT)
 	    "",
 #else
@@ -2395,9 +2510,11 @@ int gforth_args(int argc, char ** argv, char ** path, char ** imagename)
       {"ignore-async-signals", no_argument, &ignore_async_signals, 1},
       {"no-super", no_argument, &no_super, 1},
       {"no-dynamic", no_argument, &no_dynamic, 1},
+      {"no-dynamic-image", no_argument, &no_dynamic_image, 1},
       {"no-0rc", no_argument, &no_rc0, 1},
       {"dynamic", no_argument, &no_dynamic, 0},
       {"code-block-size", required_argument, NULL, opt_code_block_size},
+      {"opt-ip-updates", required_argument, NULL, opt_opt_ip_updates},
       {"print-metrics", no_argument, &print_metrics, 1},
       {"print-prims", no_argument, &print_prims, 1},
       {"print-sequences", no_argument, &print_sequences, 1},
@@ -2438,6 +2555,7 @@ int gforth_args(int argc, char ** argv, char ** path, char ** imagename)
     case 'D': print_diag(); break;
     case 'v': fputs(PACKAGE_STRING" "ARCH"\n", stderr); exit(0);
     case opt_code_block_size: if((code_area_size = convsize(optarg,sizeof(Char)))==-1L) return 1; break;
+    case opt_opt_ip_updates: opt_ip_updates = atoi(optarg); break;
     case ss_number: static_super_number = atoi(optarg); break;
     case ss_states: maxstates = max(min(atoi(optarg),MAX_STATE),1); break;
 #ifndef NO_DYNAMIC
@@ -2468,10 +2586,12 @@ Engine Options:\n\
   -m SIZE, --dictionary-size=SIZE   Specify Forth dictionary size\n\
   --map-32bit			    Try to put the dictionary in the first 2GB\n\
   --no-dynamic			    Use only statically compiled primitives\n\
+  --no-dynamic-image		    Use statically compiled primitives in image\n\
   --no-offset-im		    Load image at normal position\n\
   --no-super			    No dynamically formed superinstructions\n\
   --no-0rc			    do not load ~/.config/gforthrc0\n\
   --offset-image		    Load image at a different position\n\
+  --opt-ip-updates=n                ip-update optimization (0=disabled)\n\
   -p PATH, --path=PATH		    Search path for finding image and sources\n\
   --print-metrics		    Print some code generation metrics on exit\n\
   --print-prims			    Print primitives with usage counts on exit\n\
@@ -2548,6 +2668,11 @@ void gforth_printmetrics()
     fprintf(stderr,"lb_newstate_new = %ld\n", lb_newstate_new);
     fprintf(stderr,"lb_applicable_base_rules = %ld\n", lb_applicable_base_rules);
     fprintf(stderr,"lb_applicable_chain_rules = %ld\n", lb_applicable_chain_rules);
+    if (ip_update_metrics==NULL)
+      fprintf(stderr, "no ip update metrics recorded\n");
+    else
+      for (i=1; i<nip_update_metrics; i++)
+        fprintf(stderr,"%5ld ip update by %2d cells\n",ip_update_metrics[i],i);
   }
   if (tpa_trace) {
     fprintf(stderr, "%ld %ld lb_states\n", lb_labeler_steps, lb_newstate_new);
@@ -2739,14 +2864,16 @@ Xt gforth_find(Char * name)
 
 Cell winch_addr=0;
 
-void gforth_bootmessage()
+Cell gforth_bootmessage()
 {
+  Cell retval = -13;
   Xt bootmessage=gforth_find((Char*)"bootmessage");
   if(bootmessage != 0) {
-    gforth_execute(bootmessage);
+    retval = gforth_execute(bootmessage);
   } else {
     debugp(stderr, "Can't print bootmessage\n");
   }
+  return retval;
 }
 
 Cell gforth_start(int argc, char ** argv)
@@ -2773,9 +2900,10 @@ Cell gforth_main(int argc, char **argv, char **env)
   Cell retvalue=gforth_start(argc, argv);
   debugp(stderr, "Start returned %ld\n", retvalue);
 
-  while(retvalue == -56) { /* throw-code for quit */
-    gforth_bootmessage();
-    retvalue = gforth_quit();
+  if(retvalue == -56) { /* throw-code for quit */
+    retvalue = gforth_bootmessage();
+    if(retvalue == -56)
+      retvalue = gforth_quit();
   }
   gforth_cleanup();
   gforth_printmetrics();
