@@ -156,6 +156,8 @@ signal_throw(int sig)
     sigprocmask(SIG_SETMASK, &emptyset, NULL);
   }
 #endif
+  if(*throw_jmp_handler == NULL)
+    graceful_exit(0x80|sig);
   throw(code);
 }
 
@@ -432,6 +434,9 @@ void install_signal_handlers(void)
 #endif
 #ifdef SIGTRAP
     SIGTRAP,
+#endif
+#ifdef SIGABRT
+    SIGABRT,
 #endif
   };
 
@@ -442,9 +447,6 @@ void install_signal_handlers(void)
 #ifdef SIGHUP
     SIGHUP,
 #endif
-#ifdef SIGABRT
-    SIGABRT,
-#endif
 #ifdef SIGTERM
     SIGTERM,
 #endif
@@ -505,6 +507,9 @@ void install_signal_handlers(void)
 #ifdef SIGTRAP
   install_signal_handler(SIGTRAP, sigaction_throw);
 #endif
+#ifdef SIGABRT
+  install_signal_handler(SIGABRT, sigaction_throw);
+#endif
 #endif /* defined(SA_SIGINFO) */
 #ifdef SIGCONT
     bsd_signal(SIGCONT, termprep);
