@@ -1152,3 +1152,64 @@ DCell fmdiv (DCell num, Cell denom)
   return res;
 }
 #endif
+
+DCell utf8_fetch_plus(Char * c_addr, UCell len)
+/* Read utf8 and increment */
+{
+  uint8_t lens[16] = { 1, 1, 1, 1,
+		       1, 1, 1, 1,
+		       1, 1, 1, 1, /* Really all invalid */
+		       2, 2, 3, 4 };
+  UTetrabyte masks[5] = { 0,
+			  0x7F000000,
+			  0x1F3F0000,
+			  0x0F3F3F00,
+			  0x073F3F3F };
+  UTetrabyte vmasks[5] = { 0,
+			   0x80000000,
+			   0xE0C00000,
+			   0xF0C0C000,
+			   0xF8C0C0C0 };
+  UTetrabyte valids[5] = { 0,
+			   0x00000000,
+			   0xC0800000,
+			   0xE0808000,
+			   0xF0808080 };
+  UTetrabyte codepoint;
+  if(len<4) {
+    UTetrabyte tmp = 0;
+    memmove((Char*)&tmp, c_addr, len);
+    codepoint = tmp;
+  } else {
+    memmove((Char*)&codepoint, c_addr, 4);
+  }
+#ifndef WORDS_BIGENDIAN
+  codepoint=BSWAP32(codepoint);
+#endif
+  {
+    int cplen=lens[codepoint>>28];
+    int mask=masks[cplen];
+    int vmask=vmasks[cplen];
+    int valid=valids[cplen];
+    DCell result;
+    
+    if((codepoint & vmask) == valid) {
+      codepoint = (codepoint & mask) >> ((4-cplen) << 3);
+      codepoint =
+	((codepoint & 0x000000FF) >> 0) |
+	((codepoint & 0x0000FF00) >> 2) |
+	((codepoint & 0x00FF0000) >> 4) |
+	((codepoint & 0xFF000000) >> 6);
+      if(codepoint > 0x10FFFF) codepoint = 0xFFFD;
+    } else {
+      switch(cplen) {
+      case 4: if((codepoint &     0xC0) !=     0x80) cplen = 3;
+      case 3: if((codepoint &   0xC000) !=   0x8000) cplen = 2;
+      case 2: if((codepoint & 0xC00000) != 0x800000) cplen = 1;
+      }
+      codepoint = 0xFFFD;
+    }
+    vm_twoCell2d((Cell)(cplen),(Cell)codepoint,result);
+    return result;
+  }
+}
